local M = {}

local ClassPlayer            = require("scenario/intersection/classes/vehicles/classPlayer")
local ClassVector            = require("scenario/intersection/classes/classVector")
local fileHandler            = require("scenario/intersection/handlers/fileHandler")
local errorHandler           = require("scenario/intersection/handlers/errorHandler")
local scenarioDetailsHandler = require("scenario/intersection/handlers/scenarioDetails")

----------------------------------------------------------------------------------------------------

local teleportData = {
    destination = "",
    angle = 0,
    starting = false,
    time = 3,
    rgba = {r = 0, g = 0, b = 0, a = 0},
    pos = vec3()
}
local display = {scenario = "", displaying = false}

local saveFileData
local triggersGroup

local setup = false
local loadingScenario = false

local sourceFiles = {}
local vehicles = {player = nil}

----------------------------------------------------------------------------------------------------

local function lockScenario(triggerName, pos)
    --[[
    Positions a barrier to a scenario to prevent entering
    Parameters:
      triggerName - the name of the trigger this shield is for
      pos         - the position of the waypoint
    ]]--
    errorHandler.assertNil(triggerName, pos)

    local shieldName = triggerName.."_shield"

    -- Don't overwrite
    if not scenetree.findObject(shieldName) then
        TorqueScript.eval([[
            new TSStatic(]]..shieldName..[[) {
              shapeName = "levels/smallgrid/art/gravitationalRacing/hubWorld/scenario_shield.dae";
              position = "]]..pos.x..[[ ]]..pos.y..[[ ]]..(pos.z-3)..[[";
              scale = "6 6 6";
              collisionType = "Visible Mesh Final";
              decalType = "Visible Mesh Final";
            };
        ]])
    end
end

local function highlightScenario(show, triggerName, pos, colour)
    --[[
    Shows a waypoint over the scenario
    Parameters:
      show        - whether or not to show the waypoint
      triggerName - the name of the trigger this waypoint is for
      pos         - the position of the waypoint
      colour      - the colour of the waypoint
    ]]--
    errorHandler.assertNil(triggerName)

    local waypointName = triggerName.."_waypoint"

    -- Don't overwrite
    if not scenetree.findObject(waypointName) and show then
        -- Specific parameters for waypoint creation
        errorHandler.assertNil(pos, colour)

        TorqueScript.eval([[
            new TSStatic(]]..waypointName..[[){
              shapeName = "art/shapes/interface/checkpoint_marker.dae";
              position = "]]..pos.x..[[ ]]..pos.y..[[ ]]..(pos.z-1)..[[";
              scale = "10 10 75";
              instanceColor = "]]..colour..[[";
            };
        ]])
    else
        local visibility = not show and 1 or 0
        TorqueScript.eval(waypointName..'.hidden = "'..visibility..'";')
    end
end

local function stringToTrigger(string)
    --[[
    Formats a string to a trigger name by forcing lower case and replacing spaces with underscores
    Parameters:
      string - the string to convert
    Returns:
      <string> - the converted string
    ]]--
    errorHandler.assertNil(string)
    return string:lower():gsub(" ", "_")
end

local function scenarioNameToTrigger(scName)
    --[[
    Converts a formatted scenario name to its trigger name
    Parameters:
      scName - the formatted scenario name
    Returns:
      <string> - the trigger name
    ]]--
    errorHandler.assertNil(scName)
    return stringToTrigger("sc_"..scName)
end

local function triggerNameFormatted(trigName)
    --[[
    Converts a trigger name to a formatted scenario/championship name
    Parameters:
      trigName - the name of the trigger to convert
    Returns:
      convertedName - the converted name
    ]]--
    errorHandler.assertNil(trigName)
    -- Remove prefix
    trigName = trigName:gsub("sc_", ""):gsub("ch_", "")
    -- Capitalise 1st letter
    trigName = trigName:sub(1, 1):upper()..trigName:sub(2, #trigName)

    local convertedName = ""
    local i = 1

    while i <= #trigName do
        local char = trigName:sub(i, i)

        if char == "_" then
            -- First character cannot start with a blank space
            char = i == 1 and "" or " "

            i = i+1
            -- Add in a blank or empty space and capitalise the next char
            convertedName = convertedName ..char..trigName:sub(i, i):upper()
        else
            -- Add in next char
            convertedName = convertedName ..char
        end

        i = i+1
    end

    return convertedName
end

local function setupScenarios()
    --[[
    Sets up the scenarios by locking those which are not available to the player,
    placing the medals achieved and indicating new scenarios which have been unlocked
    ]]--
    local scenarios = saveFileData.scenarios
    for scName, scData in pairs(scenarios) do
        local triggerName = scenarioNameToTrigger(scName)
        local trigger = scenetree.findObject(triggerName)

        if trigger then
            local pos = trigger:getPosition()

            local fileName = sourceFiles[triggerName:gsub("sc_", "")]

            if not scData.unlocked then
                lockScenario(triggerName, pos)
            elseif scData.score == 0 then
                local _, difficulty = scenarioDetailsHandler.getScenarioDetails(fileName)
                local difficultyColour = scenarioDetailsHandler.difficultyToColourRGBA(difficulty)
                highlightScenario(true, triggerName, pos, difficultyColour)
            end
        end
    end
end

local function setupSourceFiles()
    --[[
    Creates the mappings of scenario name to source file
    ]]--
    for _, objName in ipairs(triggersGroup) do
        local scenarioName = objName:gsub("sc_", "")
        sourceFiles[scenarioName] = scenarioDetailsHandler.getSourcePath(scenarioName)
    end
end

local function showScenarioPreview(fileName)
    --[[
    Shows the scenario preview on the UI for a given scenario
    Parameters:
        fileName - the scenario file name
    ]]--
    errorHandler.assertNil(fileName)

    local sourceFile = scenarioDetailsHandler.getSourcePath(fileName)
    local scFile = jsonDecode(readFile(sourceFile), sourceFile)[1]
    local scTitle, scDifficulty = scenarioDetailsHandler.getScenarioDetailsByFilename(fileName)

    local saveData = saveFileData.scenarios[fileName]

    local data = {
        scTitle = scTitle,
        scDifficulty = scDifficulty,
        scImage = fileName..".png",
        best = {
            score = saveData.score,
            medal = saveData.medal
        },
        scores = scFile.intersection.medalRequirement,
        levelSettings = {
            weather = scFile.intersection.weather,
            time = scFile.intersection.ToD
        },
        location = scFile.intersection.location
    }

    guihooks.trigger("i_scenario_preview", data)
end

--------------------------------------------------------------------------------------------------------------------
-- Teleport Functions------------------------------------------------------------------------------------------------
--------------------------------------------------------------------------------------------------------------------

local function teleportToScenario(scenario)
    --[[
    Loads the specified scenario
    Parameters:
      scenario     - the scenario to load
    ]]--
    errorHandler.assertNil(scenario)

    local filePath = sourceFiles[scenario]

    scenario_scenariosLoader.start(scenario_scenariosLoader.loadScenario(filePath, nil, filePath))
    -- Set to prevent continual loading of this scenario
    loadingScenario = true
end

local function beginTeleport(dt)
    --[[
    Begins the teleport
    Parameters:
      dt - the time since last frame
    ]]--
    errorHandler.assertNil(dt)

    local pos = teleportData.pos

    local rgba = teleportData.rgba
    rgba.a = 1 - teleportData.time/3

    for i = 1, 3 do
        TorqueScript.eval('teleport'..i..'.position = "'..pos.x..' '..pos.y..' '..pos.z..'";')
        TorqueScript.eval('teleport'..i..'.instanceColor = "'..rgba.r..' '..rgba.g..' '..rgba.b..' '..rgba.a..'";')
    end

    teleportData.time = teleportData.time - dt

    if teleportData.time <= 0 then
        teleportToScenario(teleportData.destination)
        -- Prevents this function being called while the new scenario is being loaded
        teleportData.starting = false
    end
end

local function cancelTeleport()
    --[[
    Cancels the teleport by resetting the teleport data
    ]]--
    teleportData.destination = ""
    teleportData.starting = false
    teleportData.time = 3

    for i = 1, 3 do
        TorqueScript.eval('teleport'..i..'.position = "10000 0 0";')
        TorqueScript.eval('teleport'..i..'.instanceColor = "0 0 1 0";')
    end
end

-------------------------------------------------------------------------------------------------------------------------------------------------------

local function onPreRender(dt)
    if setup and not loadingScenario then
        vehicles.scenario_player0:update(dt)

        -- Begin the teleporting procedure
        if teleportData.starting then
            beginTeleport(dt)
        end

        local vehPos = vehicles.scenario_player0:getPosition()

        if triggersGroup then
            if not display.displaying then
                -- Check each trigger to see if the player is close by
                for _, objName in ipairs(triggersGroup) do
                    local objPos = scenetree.findObject(objName):getPosition()

                    local dist = math.sqrt((vehPos:getX()-objPos.x)^2 + (vehPos:getY()-objPos.y)^2 + (vehPos:getZ()-objPos.z)^2)

                    if dist <= 10 then
                        showScenarioPreview(objName:gsub("sc_", ""))

                        -- Prevent unnecessary looping of this code with data is being displayed
                        display.displaying = true
                        display.scenario = objName
                    end
                end
            else
                local objPos = scenetree.findObject(display.scenario):getPosition()
                local dist = math.sqrt((vehPos:getX()-objPos.x)^2 + (vehPos:getY()-objPos.y)^2 + (vehPos:getZ()-objPos.z)^2)
                if dist > 10 then
                    -- Vehicle has moved out of the area so the UI should be reset
                    guihooks.trigger("i_scenario_preview_hide")
                    -- Reset variables
                    display.displaying = false
                    display.scenario = ""
                end
            end
        end
    end
end

local function onBeamNGTrigger(data)
    if data.event == "enter" then
        ui_message("Teleporting...", 3, "A")
        teleportData.starting = true
        teleportData.destination = data.triggerName:gsub("sc_", "")

        -- Get the rgba for the teleport object
        local _, dif = scenarioDetailsHandler.getScenarioDetails(sourceFiles[teleportData.destination])
        local rgbaString = scenarioDetailsHandler.difficultyToColourRGBA(dif)
        local rgba = {}
        -- Match decimals or integers
        for val in rgbaString:gmatch("(%d*%.?%d+)") do
            table.insert(rgba, val)
        end

        teleportData.rgba = {r = rgba[1], g = rgba[2], b = rgba[3], a = 0}

        -- Place markers
        teleportData.pos = scenetree.findObject(data.triggerName):getPosition()
    else
        ui_message("Teleporting cancelled", 3, "A")
        -- Reset data back to initial state
        cancelTeleport()
    end
end

--local function generateForest()
--    --[[
--    Generates a forest for the level
--    ]]--
--    local forestGenerator = require("scenario/intersection/forestGenerator")
--
--    local startPos = ClassVector:new(-86, 5, 46.7)
--    forestGenerator.generateTrees(
--            startPos:getZ(),
--            startPos,
--            ClassVector:new(-22, 76, 0),
--            0.75,
--            15,
--            ClassVector:new(3, 3, 3),
--            2,
--            "scots_pine"
--    )
--end

local function setupSideInfo()
    --[[
    Setups the side UI
    ]]--
    guihooks.trigger("i_display_stats", fileHandler.getNumberOfEachMedals())
end

local function onScenarioChange(sc)
    if sc then
        if sc.state == "pre-running" then
            -- Hide the preview UI
            guihooks.trigger("i_scenario_preview_hide")

            -- Reset teleport objects
            for i = 1, 3 do
                TorqueScript.eval('teleport'..i..'.position = "10000 0 0";')
            end

            if setup then
                return
            end

            if not vehicles.scenario_player0 then
                local playerObj = scenetree.findObject("scenario_player0")
                local pos = playerObj:getPosition()
                local playerPos = ClassVector:new(pos.x, pos.y, pos.z)

                vehicles.scenario_player0 = ClassPlayer:new(playerObj:getID(), "scenario_player0", playerPos)
            end

            fileHandler.checkForUpdate()

            saveFileData = fileHandler.readWholeFile()

            triggersGroup = scenetree.findClassObjects('BeamNGTrigger')

            setupSourceFiles()

            setupScenarios()
            be:reloadStaticCollision()

            -- Hide the floor
            local ground = scenetree.findClassObjects("Groundplane")
            for _, name in ipairs(ground) do
                scenetree.findObject(name).hidden = true
            end

            setup = true

            -- generateForest()
            --require("scenario/intersection/handlers/forestHandler").shuffleTrees("forest_mix")
        elseif sc.state == "running" then
            guihooks.trigger("i_scenario_preview_hide")
            setupSideInfo()
            -- Force a quick start - don't wait for countdown
            sc.countDownTime = 0
        end
    end
end

M.onPreRender = onPreRender
M.onBeamNGTrigger = onBeamNGTrigger
M.onScenarioChange = onScenarioChange
return M
