local errorHandler = require("scenario/intersection/handlers/errorHandler")

local function getDefaultClass()
    --[[
    Returns:
        <class> - the default structure for a class, that is, a class with two basic, intrinsic
                  methods: instanceOf, findInstancesOf
    ]]--
    local newClass = {}

    function newClass:instanceOf(className)
        --[[
        Returns whether an object is a instance of a class, mimicking instanceof
        Permits inheritance ie. upcasting
        Parameters:
            object    - the object
            className - the name of the class
        Returns:
            <boolean> - whether the object is an instance of a class
        ]]--
        errorHandler.assertNil(className)

        if not self.instanceOf then
            return false
        end

        if self.className == className then
            return true
        end

        -- Try the next level up the hierarchy, if there is one
        local metatable = getmetatable(self)
        if metatable then
            return self.instanceOf(metatable.__index, className)
        end

        return false
    end

    function newClass:findInstancesOf(instances, className)
        --[[
        Finds and returns the instances which are of a sp[ecific subclass
        Parameters:
            instances - the instances to check
            className - the baseclass name to filter with
        Returns:
            <table <className>> - the instances which match
        ]]--
        errorHandler.assertNil(instances, className)

        local matching = {}

        for _, instance in pairs(instances) do
            if instance:instanceOf(className) then
                table.insert(matching, instance)
            end
        end

        return matching
    end

    return newClass
end

--local function inheritsFromMulti(baseClass1, baseClass2)
--    --[[
--    Mimics double inheritance for Lua 'classes'
--    Parameters:
--        baseClass1 - the first class to inherit from
--        baseClass2 - the second class to inherit from
--    Returns:
--        <class> - the class object
--    ]]--
--    local newClass = getDefaultClass()
--
--    setmetatable(newClass, { __index = function(_, k)
--        for i = 1, #arg do
--            local isParent = arg[i][k]
--            if isParent then
--                return isParent
--            end
--        end
--    end})
--
--    newClass.__index = newClass
--
--    return newClass
--end

local function inheritsFrom(baseClass)
    --[[
    Mimics inheritance for Lua 'classes'
    Parameters:
    `   baseClass - the class this class will inherit from (can be nil)
    Returns:
        <class> - the class object
    ]]--
    local newClass = getDefaultClass()

    -- Can inherit from nil
    if baseClass then
        setmetatable(newClass, { __index = baseClass})
    end

    return newClass
end

return {
    inheritsFrom = inheritsFrom
}