------------------------------------------------------------------------------------------------------------------
-- Import
----------------------------------------------------------------------------------------------------

local classHelper  = require("scenario/intersection/classHelper")
local errorHandler = require("scenario/intersection/handlers/errorHandler")
local soundHandler = require("scenario/intersection/handlers/soundHandler")
local crashbreaker = require("scenario/intersection/handlers/crashbreaker")

local scoringHandler = require("scenario/intersection/handlers/scoringHandler")

local ClassShockwave = require("scenario/intersection/classes/classShockwave")

------------------------------------------------------------------------------------------------------------------

--[[
NOTE: This class is implied as an ABSTRACT, so no instances should be created!
]]--
local ClassStandardExplosion = classHelper.inheritsFrom(nil)

-------------------------------------------------------------------------------------------------------
-- Static Attributes
----------------------------------------------------------------------------------------------------

ClassStandardExplosion.nextID = 1

------------------------------------------------------------------------------------------------------------------
-- Instance Methods
----------------------------------------------------------------------------------------------------

function ClassStandardExplosion:update(dt)
    --[[
    Updates the explosion effect
    Parameters:
        dt - the time since last frame
    Returns:
        <boolean> - whether the explosion is finished
    ]]--
    local data = self.useData
    local constants = self.constants

    data.timer = data.timer + dt

    -- The scale of the outer cloud
    local baseScale = constants.EXPLOSION_RANGE * (data.timer/constants.PERIOD)
    -- Fade out alpha as timer increases
    self.shockwave:updateShockwave(baseScale, math.max(0, (constants.PERIOD - data.timer) * constants.MAX_ALPHA))

    -- Affect the vehicles
    for k, instance in pairs(self.vehicles) do
        -- Do not affect the vehicle who triggered the explosion and those already under the effect
        if k ~= data.vehicleTrigger and not self.vehiclesAffected[k] then
            local distVec = instance:getPosition():subtract(data.position)
            local dist = distVec:getMagnitude()

            if dist <= baseScale/2 then
                -- Apply a force of max magnitude 300 outwards for 1/4 of a second
                instance:boost(300 * (1-dist/constants.EXPLOSION_RANGE/2), data.position, 0.25)
                scoringHandler.addExplosion(instance)
                instance:setCrashed()
                instance:lock()

                self.vehiclesAffected[k] = true

                local cb = crashbreaker.getCrashbreaker()
                -- Only vehicles count
                if cb and not instance:instanceOf("ClassProp") then
                    cb:addCrashedVehicle()
                end
            end
        end
    end

    -- Timer is up
    if data.timer >= constants.PERIOD then
        self.shockwave:reset()
        return true
    end

    return false
end

function ClassStandardExplosion:instanceOf()
    return "ClassStandardExplosion"
end

function ClassStandardExplosion:use(vehicle)
    --[[
    Begins the usage sequence
    Parameters:
        vehicle - the vehicle triggering the explosion
    ]]--
    errorHandler.assertNil(vehicle)

    --[[Set the explosive's position at the vehicle's
        Since position is a table, immutability requires a cloned vector]]--
    self.useData.position = vehicle:getPosition():clone()
    self.useData.vehicleTrigger = vehicle:getId()

    soundHandler.play("crashbreaker_explosion")

    self.shockwave:begin(self.useData.position)

    -- Destroy the vehicle
    vehicle:breakAll()
end

function ClassStandardExplosion:reset()
    --[[
    Resets the instance back to the initial state
    ]]--
    self.shockwave:reset()

    self.useData = {
        timer = 0,
        vehicleTrigger = nil,
        position = nil
    }

    for k, _  in pairs(self.vehicles) do
        self.vehiclesAffected[k] = false
    end
end

-------------------------------------------------------------------------------------------------------
-- Constructor
----------------------------------------------------------------------------------------------------

function ClassStandardExplosion:new(vehicles)
    --[[
    Parameters:
        vehicles - the list of vehicles that can be affected by this explosion
    ]]--
    local self = {}

    self.id = ClassStandardExplosion.nextID
    ClassStandardExplosion.nextID = ClassStandardExplosion.nextID + 1

    self.vehicles = vehicles

    -- Stores whether a vehicle has been affected
    self.vehiclesAffected = {}
    for k, _  in pairs(self.vehicles) do
        self.vehiclesAffected[k] = false
    end

    self.constants = {
        -- The diameter the explosive affects other objects
        EXPLOSION_RANGE = 40,
        -- The total time the explosive lasts
        PERIOD = 1,
        -- The maximum alpha of the shockwaves
        MAX_ALPHA = 1,
        -- The number of shockwaves
        NUM_SHOCKWAVES = 1
    }

    self.useData = {
        timer = 0,
        position = nil,
        vehicleTrigger = nil
    }

    self.shockwave = ClassShockwave:new(
            "explosion_shockwave_"..self.id,
            self.constants.NUM_SHOCKWAVES,
            self.constants.EXPLOSION_RANGE,
            self.constants.PERIOD,
            function(_, alpha)
                return "1 0.52 0 "..alpha
            end,
            true,
            nil
    )

    setmetatable(self, {__index = ClassStandardExplosion})
    return self
end

------------------------------------------------------------------------------------------------------------------
-- Interface
----------------------------------------------------------------------------------------------------

return ClassStandardExplosion