local M = {}

local errorHandler = require("scenario/intersection/handlers//errorHandler")

local vanillaMaps = {
	"automation_test_track",
	"Cliff",
	"driver_training",
	"east_coast_usa",
	"glow_city",
	"GridMap",
	"hirochi_raceway",
	"industrial",
	"italy",
	"jungle_rock_island",
	"port",
	"smallgrid",
	"small_island",
	"Utah",
	"west_coast_usa"
}

local function getSourcePath(scenarioName)
  --[[
  Gets the file path for a scenario
  Parameters:
    scenarioName - the name of the scenario
  Returns:
    <string> - the file directory of the scenario
  ]]--
  errorHandler.assertNil(scenarioName)

  for _, map in ipairs(vanillaMaps) do
    local fileDirectory = "levels/"..map.."/scenarios/intersection/crashes/"..scenarioName..".json"

    if readFile(fileDirectory) then
      return fileDirectory
    end
  end

  log("E", "getSourcePath()", "Cannot find "..(scenarioName or "nil").." as a source file!")
end

local function getVanillaMaps()
  return vanillaMaps
end

local function difficultyToColourRGBA(difficulty)
  --[[
  Gets the difficulty colour (rgba)
  Parameters:
    difficulty - the difficult level
  Returns:
    <string> - the RGBA colour
  ]]--
  errorHandler.assertNil(difficulty)

  difficulty = difficulty:lower()
  if difficulty == "easy" then
    return "0 1 0 1"
  elseif difficulty == "medium" then
    return "1 1 0 1"
  elseif difficulty == "hard" then
    return "1 0 0 1"
  else
    return "1 1 1 1"
  end
end

local function difficultyToColour(difficulty)
  --[[
  Returns
    <string> - the colour of a difficulty
  ]]--
  difficulty = difficulty:upper()

  if difficulty == "EASY" then
    return "white"
  elseif difficulty == "MEDIUM" then
    return "gold"
  elseif difficulty == "HARD" then
    return "red"
  end
end

local function getScenarioDetails(sourceFile)
  --[[
  Returns the title and difficulty of the scenario
  Parameters:
    sourceFile - the file directory of the scenario
  Returns:
    <string> - the name of the scenario
    <string> - the difficulty of the scenario
  ]]--
  errorHandler.assertNil(sourceFile)

  local scenarioTitle = jsonDecode(readFile(sourceFile), sourceFile)[1].name
  -- Matches the first part of the title
  local scenarioName = scenarioTitle:match("[%a%s']*")
  -- Remove space at the end of the name and the start of the difficulty
  scenarioName = scenarioName:sub(1, #scenarioName-1)
  -- Matches the seconds part in the brackets
  local scenarioDifficulty = scenarioTitle:match("%((%a+)%)")

  return scenarioName, scenarioDifficulty
end

local function getScenarioDetailsByFilename(fileName)
    --[[
    Gets the difficult and title of the scenario from the scenario file name
    Parameters:
        fileName - the name of the scenario file
    Returns
        <string> - the scenario title
        <string> - the difficulty level
    ]]--
    return getScenarioDetails(getSourcePath(fileName))
end

local function getSpecificDetail(attribute, sourceFile)
  --[[
  Returns a specific attribute of a scenario
  Parameters:
    attribute  - the attribute
    sourceFile - the file directory of the scenario
  Returns:
    <?> - the attribute value
  ]]--
  errorHandler.assertNil(attribute, sourceFile)
  return jsonDecode(readFile(sourceFile), sourceFile)[1][attribute]
end

local function getScenariosSortedByDif()
	--[[
	Returns:
	    scenarios - all scenarios present, ordered into a table by difficulty
	]]--
	local scenarios = {easy = {}, medium = {}, hard = {}}

	for _, map in ipairs(vanillaMaps) do
		local path = "/levels/"..map.."/scenarios/intersection/crashes"
		local files = FS:findFilesByPattern(path, "*.json", 1, true, false)

		for _, fileName in ipairs(files) do
			local scName, scDif = getScenarioDetails(fileName)
			table.insert(scenarios[scDif:lower()], scName)
		end
	end

	return scenarios
end

M.getSourcePath = getSourcePath
M.getVanillaMaps = getVanillaMaps
M.difficultyToColourRGBA = difficultyToColourRGBA
M.difficultyToColour = difficultyToColour
M.getScenarioDetails = getScenarioDetails
M.getScenarioDetailsByFilename = getScenarioDetailsByFilename
M.getSpecificDetail = getSpecificDetail
M.getScenariosSortedByDif = getScenariosSortedByDif
return M
