local M = {}

----------------------------------------------------------------------------------------------------
-- Imports
----------------------------------------------------------------------------------------------------

local ClassVehicle = require("scenario/intersection/classes/vehicles/classVehicle")

local errorHandler = require("scenario/intersection/handlers/errorHandler")
local tableComp    = require("scenario/intersection/tableComprehension")

----------------------------------------------------------------------------------------------------
-- Variables
----------------------------------------------------------------------------------------------------

local constants = {
    bonuses = {
        -- The score for a vehicle becoming involved in the crash (only once)
        INVOLVEMENT = 2500,
        TARGET_INVOLVEMENT = 15000,
        -- The multiplier to add for being involved
        ADD_MULTI = 0.1,
    },
    explosions = {
        -- The score for a vehicle being impacted by an explosion (can be more than once)
        INVOLVEMENT = 2500,
        -- The multiplier to add for being impacted in an explosion
        ADD_MULTI   = 0.2,
    },
    -- The ratio of 1 ms^-1:x damage
    velocityScoreRatio = 100
}

local scoring = {
    multiplier = 1,
    breakdown = {
        explosions = 0,
        vehicleCrashed = 0,
        collisions = 0,
        damage = 0
    }
}

----------------------------------------------------------------------------------------------------

local function getConstants()
    --[[
    Returns:
        <table> - the constants for scoring
    ]]--
    return constants
end

local function getPointsBreakdown()
    --[[
    Returns:
        <table> - the breakdown of points accumulated
    ]]--
    return scoring
end

local function getTotalScore()
    --[[
    Returns:
        <number> - the total score
    ]]--
    local total = 0

    for _, score in pairs(scoring.breakdown) do
        total = total + score
    end

    return total
end

local function formatToUI(coins, multipliers, vehicles)
    --[[
    Formats the scoring for use of the end screen
    Parameters:
        coins       - the coin instances in the scenario
        multipliers - the multiplier instances in the scenario
        vehicles    - the vehicle instances in the scenario
    Returns:
        <table> - the formatted table
        <table> - the multiplier data (may be nil)
        <table> - the target data (may be nil)
    ]]--
    errorHandler.assertNil(coins, multipliers, vehicles)

    local breakdown = scoring.breakdown

    local coinTotal = 0
    local coinsRow  = {}

    -- Sort out the coins and which coins to add in
    for k, instance in pairs(coins) do
        if instance:isCollected() then
            local value = instance:getValue(k)

            coinTotal = coinTotal + value

            table.insert(coinsRow, {type = k, value = value})
        end
    end


    local numAis = tableComp.lengthOfTable(vehicles.ais)
    local max = {
        damage = 0,
        involvement = numAis * constants.bonuses.INVOLVEMENT,
        explosions = numAis * constants.explosions.INVOLVEMENT
    }

    -- Sum up the vehicle maximum amount
    for _, instance in pairs(vehicles.ais) do
        max.damage = max.damage + instance.MAX_DAMAGE_SCORE
    end


    -- Will only be at most one target vehicle in a scenario
    local targetAI = ClassVehicle:findInstancesOf(vehicles.ais, "ClassTarget")[1]
    -- May not exist so not necessary for the UI to know about it
    local target
    if targetAI then
        target = {value = constants.bonuses.TARGET_INVOLVEMENT, gotten = targetAI:hasCrashed()}
    end


    local data = {
        {reason = "Collisions"    , amount = breakdown.collisions    , max = nil},
        {reason = "Vehicles"      , amount = breakdown.vehicleCrashed, max = max.involvement},
        {reason = "Explosions"    , amount = breakdown.explosions    , max = max.explosions},
        {reason = "Vehicle Damage", amount = breakdown.damage        , max = max.damage},
        {reason = "Coins"         , amount = coinsRow}
    }


    -- A scenario with no multipliers has a nil multiplier row so the UI can ignore it
    local multiplierRow
    local multiplier = 1

    if tableComp.lengthOfTable(multipliers) > 0 then
        multiplierRow = {}

        -- Sort out the multipliers and which multipliers to add in
        for k, instance in pairs(multipliers) do
            if instance:isCollected() then
                local value = instance:getValue()

                -- Half-multiplier has the highest precedence, followed by value
                if multiplier ~= 0.5 and (value > multiplier or value == 0.5) then
                    multiplier = value
                end

                table.insert(multiplierRow, {type = k, value = value})
            end
        end

        if #multiplierRow > 0 then
            -- Sort based on order of [x2, x4, xh]
            table.sort(multiplierRow, function(a,b)
                if a.value == 0.5 then return false end

                if b.value == 0.5 or a.value < b.value then
                    return true
                end
            end)

        else
            -- Re-initialise to nil so the UI skips it as no multipliers were picked up
            multiplierRow = nil
        end
    end

    -- Base damage
    local total = getTotalScore()
    -- Add on multiplier (only affects base damage, not bonuses)
    total = total * multiplier
    -- Add on coins
    total = total + coinTotal

    if target and target.gotten then
        -- Add on target vehicle
        total = total + target.value
    end

    -- Set total
    table.insert(data, {amount = total})

    return data, multiplierRow, target
end

--local function isABetterMedal(oldMedal, newMedal)
--    --[[
--    Returns whether a medal is better than another
--    Parameters:
--        oldMedal - the medal to to be compared to
--        newMedal - the medal to check whether it is better than oldMedal
--    Returns:
--        <boolean> - whether newMedal is better than oldMedal
--    ]]--
--    if oldMedal == "dev" then
--        return false
--    end
--end

local function getMedalAchieved(scores, scoreAchieved)
    --[[
    Returns the medal achieved
    Parameters:
        scores - the scores for each medal
        scoreAchieved - the achieved score
    Returns:
        <string> - the medal achieved (none if none achieved)
    ]]--
    local current = {
        medal = "none",
        medalScore = 0
    }

    for k, score in pairs(scores) do
        -- If achieved and better than the score for the currently found medal, update it
        if scoreAchieved >= score and score > current.medalScore then
            current.medal = k
            current.medalScore = score
        end
    end

    return current.medal
end

local function updateScoringUI()
    --[[
    Updates the scoring UI
    ]]--
    guihooks.trigger("i_scoring_update", {score = getTotalScore(), multiplier = scoring.multiplier})
end

local function updateVehicleDamages(vehicles)
    --[[
    Updates the total vehicle damages done thus far
    Parameters:
        vehicles - the vehicles that can be damaged
    ]]--
    local vehicleDamage = 0

    for _, instance in pairs(vehicles) do
        vehicleDamage = vehicleDamage + instance:getDamageScore()
    end

    scoring.breakdown.damage = vehicleDamage
end

local function addExplosion(vehicle)
    --[[
    Adds an amount for an explosion
    Parameters:
        vehicle - the vehicle involved
    ]]--
    errorHandler.assertNil(vehicle)

    -- Ignore props
    if vehicle:instanceOf("ClassProp") then
        return
    end

    local breakdown = scoring.breakdown

    local explosionBonus = constants.explosions.INVOLVEMENT
    breakdown.explosions = breakdown.explosions + explosionBonus

    scoring.multiplier = scoring.multiplier + constants.explosions.ADD_MULTI

    -- Counts as an initial involvement
    if not vehicle:hasCrashed() then
        local crashBonus = constants.explosions.INVOLVEMENT
        breakdown.vehicleCrashed = breakdown.vehicleCrashed + crashBonus

        scoring.multiplier = scoring.multiplier + constants.bonuses.ADD_MULTI
    end
end

local function addScoreFromCollision(vehicle1, vehicle2)
    --[[
    Calculates and adds the score from a collision
    Parameters:
        vehicle1 - the first vehicle in the collision
        vehicle2 - the other vehicle in the collision
    ]]--
    errorHandler.assertNil(vehicle1, vehicle2)

    --[[To get a non-zero collision score:
        - both vehicles must not have been crashed into yet
        - vehicles cannot both be props]]--
    local areProps = vehicle1:instanceOf("ClassProp") and vehicle2:instanceOf("ClassProp")
    if (vehicle1:hasCrashed() and vehicle2:hasCrashed()) or areProps then
        return
    end

    --[[The score from the act of the collision
        Calculate the total velocity in the collision and multiplier it by the necessary factors]]--
    local collisionScore = math.floor((vehicle1:getVelocity() - vehicle2:getVelocity()):length() * constants.velocityScoreRatio * scoring.multiplier)

    -- If the vehicles have not been crashed yet, apply an involvement bonus (both multiplier and score)
    local involvementBonus = 0
    local additionalMulti = 0

    for _, inst in ipairs({vehicle1, vehicle2}) do
        -- Ignore player
        if inst:instanceOf("ClassAi") and not inst:hasCrashed() then
            involvementBonus = involvementBonus + constants.bonuses.INVOLVEMENT
            additionalMulti = additionalMulti + constants.bonuses.ADD_MULTI
        end
    end

    -- Add the multiplier first before adding the score
    scoring.multiplier = scoring.multiplier + additionalMulti

    -- Bonuses are not affected by the collision multiplier
    scoring.breakdown.vehicleCrashed = scoring.breakdown.vehicleCrashed + involvementBonus
    scoring.breakdown.collisions     = scoring.breakdown.collisions     + collisionScore
end

local function addScore(reason, amount)
    --[[
    Adds a custom amount
    ]]--
    errorHandler.assertNil(reason, amount)
    errorHandler.assertNil(scoring.breakdown[reason], reason.." is not a valid reason")

    scoring.breakdown[reason] = scoring.breakdown[reason] + amount
end

local function reset()
    --[[
    Resets the scoring variables
    ]]--
    scoring = {
        multiplier = 1,
        breakdown = {
            explosions = 0,
            vehicleCrashed = 0,
            collisions = 0,
            damage = 0
        }
    }
    updateScoringUI()
end

local function resetState()
    --[[
    Resets variables back to their initial state
    ]]--
    reset()
end

M.getConstants = getConstants
M.getPointsBreakdown = getPointsBreakdown
M.formatToUI = formatToUI
M.updateScoringUI = updateScoringUI
M.getMedalAchieved = getMedalAchieved
M.updateVehicleDamages = updateVehicleDamages
M.addExplosion = addExplosion
M.addScoreFromCollision = addScoreFromCollision
M.addScore = addScore
M.reset = reset
M.resetState = resetState
return M