angular.module('beamng.apps')
  .directive('intersectionScenarioPreview', [function () {
    return {
      templateUrl: "modules/intersection/apps/ScenarioPreviewScreen/templateBMNG.html",
      replace: true,
      link: function (scope) {
        'use strict';

        function resetValues() {
          scope.$evalAsync(function () {
          });
        }

        /**
         * Upper-cases the first letter in string
         * @param string - the string to uppercase
         * @returns {string} the upper-cased string
         */
        function upperCaseFirstLetter(string) {
          return string[0].toUpperCase() + string.slice(1);
        }

        /**
         * Capitalises all words in a string. A word can be defined using the splitBy parameter
         * @param string - the string to capitalise
         * @param splitBy - the divider that counts as a gap between words
         * @returns {string} the resultant string
         */
        function capitaliseString(string, splitBy) {
          let newString = "";

          for (let word of string.split(splitBy)) {
            newString += upperCaseFirstLetter(word) + splitBy;
          }

          // Remove last splitter
          newString = newString.slice(0, -splitBy.length);

          return newString;
        }

        /**
         * Returns the icon that reflects a time and weather setting
         * @param {string} time    - the time of day
         * @param {string} weather - the weather setting
         * @returns {string} the icon
         */
        function getWeatherIcon(time, weather) {
          if (time === "default")
            return "sun";

          // Snow is a single image icon, so time is irrelevant
          if (weather.search("snow") !== -1)
            return "snowflake";

          if (time === "morning" || time === "noon") {
            if (weather === "sunny")
              return "sun";
            else if (weather === "light-rain")
              return "cloud-sun-rain";
            else if (weather === "heavy-rain")
              return "cloud-showers-heavy";
            else
              console.warn(weather, " is an invalid weather setting!");
          } else if (time === "midnight") {
            if (weather === "sunny")
              return "moon";
            else if (weather === "light-rain")
              return "cloud-moon-rain";
            else if (weather === "heavy-rain")
              return "cloud-showers-heavy";
            else
              console.warn(weather, " is an invalid weather setting!");
          } else {
            console.warn(time, " is an invalid time setting!");
          }
        }

        /**
         * Returns the colour of a difficulty
         * @param difficulty - the difficulty to get the colour of
         * @returns {string} the difficulty colour in hex
         */
        function getDifficultyColour(difficulty) {
          difficulty = difficulty.toLowerCase();

          if (difficulty === "easy")
            return "#00FF00";
          else if (difficulty === "medium")
            return "#FFFF00";
          else if (difficulty === "hard")
            return "#FF0000";
          else
            console.warn(difficulty, " is an invalid difficulty!");
        }

        /**
         * Returns the colour of a medal
         * @param medal - the medal name
         * @returns {string} the colour of the medal
         */
        function getMedalColour(medal) {
          if (medal === "bronze")
            return "#cd7f32";
          else if (medal === "silver")
            return "silver";
          else if (medal === "gold")
            return "gold";
          else if (medal === "chaos")
            return "b9f2ff";
          else if (medal === "dev")
            return "#00ff00";
          else if (medal === "none" || medal === "")
            return "white";
          else
            console.warn(medal, " not a valid medal!");
        }

        /**
         * Fills in the preview with all the relevant scenario data
         * @param data - the scenario data, should contain:
         *               - scTitle - the scenario title
         *               - scImage - the preview image of the scenario
         *               - scores - the scores for bronze, silver, gold, chaos, dev
         *               - best - the best attempt, should include:
         *                  - score - the player score
         *                  - medal - the medal achieved
         *               - levelSettings - the time and weather. Valid options are:
         *                  - time: morning, noon, default, midnight
         *                  - weather: sunny, light-rain, light-snow, heavy-rain, heavy-snow
         */
        scope.$on('i_scenario_preview', function (event, data) {
          let imagesDir = "modules/intersection/images";

          document.getElementById("uiContainer").style.opacity = "1";


          let difColour = getDifficultyColour(data.scDifficulty);
          let title = document.getElementById("scTitle");
          title.innerHTML = data.scTitle + " (" + data.scDifficulty + ")";
          title.style.color = difColour;

          let img = document.getElementById("scImage");
          img.src = imagesDir + "/scenarios/" + data.scImage;
          img.style.border = "2px solid " + difColour;


          // Scores
          let scores = data.scores;
          for (let [key, value] of Object.entries(scores)) {
            document.getElementById("score" + upperCaseFirstLetter(key)).innerHTML = value;
          }

          // Best score
          let best = data.best;
          let element = document.getElementById("scoreBest");
          element.innerHTML = best.score;
          element.style.color = getMedalColour(best.medal);


          // Weather and time
          let levelSettings = data.levelSettings;
          levelSettings.weather = levelSettings.weather ? levelSettings.weather : "sunny";
          levelSettings.time    = levelSettings.time    ? levelSettings.time    : "noon";
          document.getElementById("weather").innerHTML = capitaliseString(levelSettings.weather, "-");
          document.getElementById("weatherIcon").src = imagesDir + "/icons/" + getWeatherIcon(levelSettings.time, levelSettings.weather) + ".svg";
          document.getElementById("time").innerHTML = upperCaseFirstLetter(levelSettings.time);

          // Location
          document.getElementById("location").innerHTML = data.location;
        });

        /**
         * Hides the UI screen
         */
        scope.$on('i_scenario_preview_hide', function() {
          document.getElementById("uiContainer").style.opacity = "0";
          document.getElementById("scImage").src = "modules/intersection/images/null.png";
        });

        scope.$on('ScenarioResetTimer', resetValues);
      }
    };
  }]);
